# =================================================================
#
# Authors: Erez Shinan, Seth Girvin
#
# Copyright (c) 2020 Seth Girvin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

"""
Module to transform an AST (Abstract Syntax Tree) to a
Python dict structure
"""

from __future__ import annotations
import logging
from collections import OrderedDict
from lark import Tree
from lark.visitors import Transformer_InPlace, Transformer, v_args
from lark.lexer import Token
from .parser import lark_cython
from typing import Any
from mappyfile.tokens import SINGLETON_COMPOSITE_NAMES, REPEATED_KEYS
from mappyfile.ordereddict import CaseInsensitiveOrderedDict
from mappyfile.quoter import Quoter


if lark_cython:
    TOKEN_TYPES = (Token, lark_cython.Token)

    def update_token_value(t, value):
        return Token.new_borrow_pos(t.type, value, t)

else:
    TOKEN_TYPES = Token  # type: ignore

    def update_token_value(t, value):
        t.value = value
        return t


log = logging.getLogger("mappyfile")


class MapfileTransformer(Transformer):
    def __init__(self, include_position: bool = False, include_comments: bool = False):
        self.quoter = Quoter()
        self.include_position = include_position
        self.include_comments = include_comments

    def key_name(self, token) -> str:
        return token.value.lower()

    def start(self, composites):
        """
        Parses a MapServer Mapfile
        Parsing of partial Mapfiles or lists of composites is also possible
        """

        # only return a list when there are multiple root composites (e.g.
        # several CLASSes)
        if len(composites) == 1:
            return composites[0]

        return composites

    def flatten(self, values: list[Any]) -> list[Any]:
        flat_list = []

        for v in values:
            if isinstance(v, TOKEN_TYPES):
                flat_list.append(v)
            elif isinstance(v, list):
                flat_list += v
            elif isinstance(v, tuple):
                flat_list += v
            elif isinstance(v, dict):
                assert "__tokens__" in v
                flat_list += v["__tokens__"]
            else:
                raise ValueError("Attribute value type not supported", v)

        return flat_list

    def plural(self, s: str) -> str:
        if s.endswith("s"):
            return s + "es"

        return s + "s"

    def create_position_dict(self, key_token, values) -> dict:
        line, column = key_token.line, key_token.column
        d = OrderedDict()
        d["line"] = line
        d["column"] = column

        if values:
            flat_list = self.flatten(values)
            value_positions = [(v.line, v.column) for v in flat_list]
            d["values"] = value_positions

        return d

    def get_single_key(self, d: dict):
        keys = list(d.keys())  # convert to list for py3
        assert len(keys) == 1
        return keys[0]

    def composite_body(self, t):
        return t

    def composite_type(self, t):
        return t

    def _process_composite_config(
        self,
        composite_dict: dict,
        attribute_dict: dict,
        position_dict: dict | None,
        pos: int,
    ):
        key_name = "config"

        # there may be several config dicts - one for each setting
        if key_name not in composite_dict:
            # create an initial OrderedDict
            composite_dict[key_name] = CaseInsensitiveOrderedDict(
                CaseInsensitiveOrderedDict
            )
        # populate the existing config dict
        cfg_dict = composite_dict[key_name]
        cfg_dict.update(attribute_dict[key_name])

        if position_dict is not None:
            if key_name not in position_dict:
                position_dict[key_name] = OrderedDict()

            subkey_name = self.get_single_key(attribute_dict[key_name])
            position_dict[key_name][subkey_name] = pos

    def _process_composite_points(
        self,
        composite_dict: dict,
        attribute_dict: dict,
        position_dict: dict | None,
        pos: int,
    ):
        key_name = "points"

        if key_name not in composite_dict:
            composite_dict[key_name] = attribute_dict[key_name]
        else:
            # if points are already in a feature then
            # allow for multipart features in a nested list
            existing_points = composite_dict[key_name]

            if calculate_depth(existing_points) == 2:
                composite_dict[key_name] = [existing_points]

            if key_name not in composite_dict:
                composite_dict[key_name] = []
            composite_dict[key_name].append(attribute_dict[key_name])

        if position_dict is not None:
            if key_name not in position_dict:
                position_dict[key_name] = pos
            else:
                existing_pos = position_dict[key_name]
                if isinstance(existing_pos, dict):
                    position_dict[key_name] = [existing_pos]
                position_dict[key_name].append(pos)

    def composite(self, t):
        """
        Handle the composite types e.g. CLASS..END
        t is a list in the form [[Token(__LAYER36, 'LAYER')], [OrderedDict([...])]]

        """
        if len(t) == 1:
            return t[0]  # metadata and values - already processed

        key_token = t[0][0]
        attribute_dicts = t[1]

        if not isinstance(attribute_dicts, list):
            # always handle a list of attributes
            attribute_dicts = [attribute_dicts]

        key_name = self.key_name(key_token)
        composite_dict = CaseInsensitiveOrderedDict(CaseInsensitiveOrderedDict)
        composite_dict["__type__"] = key_name

        if self.include_position:
            position_dict = self.create_position_dict(key_token, None)
            composite_dict["__position__"] = position_dict
        else:
            position_dict = None

        if self.include_comments:
            comments_dict = composite_dict["__comments__"] = OrderedDict()

        for d in attribute_dicts:
            keys = d.keys()
            if "__type__" in keys:
                k = d["__type__"]
                if k in SINGLETON_COMPOSITE_NAMES:
                    composite_dict[k] = d
                else:
                    plural_key = self.plural(k)
                    if plural_key not in composite_dict:
                        composite_dict[plural_key] = []

                    composite_dict[plural_key].append(d)
            else:
                #  simple attribute
                pos = d.pop("__position__")
                d.pop(
                    "__tokens__", None
                )  # tokens are no longer needed now we have the positions
                comments = d.pop("__comments__", None)

                key_name = self.get_single_key(d)

                if key_name == "config":
                    self._process_composite_config(
                        composite_dict, d, position_dict, pos
                    )
                elif key_name == "points":
                    self._process_composite_points(
                        composite_dict, d, position_dict, pos
                    )
                elif key_name in REPEATED_KEYS:
                    if key_name not in composite_dict:
                        composite_dict[key_name] = []

                    composite_dict[key_name].append(d[key_name])

                    if position_dict is not None:
                        if key_name not in position_dict:
                            position_dict[key_name] = []
                        position_dict[key_name].append(pos)
                else:
                    assert len(d.items()) == 1
                    if position_dict is not None:
                        # hoist position details to composite
                        position_dict[key_name] = pos
                    if self.include_comments and comments:
                        # hoist comments to composite
                        comments_dict[key_name] = comments

                    composite_dict[key_name] = d[key_name]

        return composite_dict

    def clean_string(self, val: str) -> str:
        return self.quoter.remove_quotes(val)

    def attr_name(self, tokens) -> str:
        t = tokens[0]
        if not isinstance(t, TOKEN_TYPES):
            #  handle ambiguities
            t = t[0]
            assert t.value.lower() in ("symbol", "style")

        return t

    def attr(self, tokens) -> dict:
        key_token = tokens[0]

        if isinstance(key_token, (list, tuple)):
            key_token = key_token[0]
            assert self.key_name(key_token) in ("style", "symbol"), self.key_name(
                key_token
            )

        key_name = self.key_name(key_token)
        value_tokens = tokens[1:]

        if isinstance(value_tokens[0], (list, tuple)):
            # for any multi-part attributes they will be lists or tuples
            # e.g. int_pair, rgb etc.
            assert len(value_tokens) == 1
            value_tokens = value_tokens[0]

        pd = self.create_position_dict(key_token, value_tokens)
        d: dict = OrderedDict()
        d["__position__"] = pd

        if len(value_tokens) > 1:
            if key_name == "config":
                assert len(value_tokens) == 2
                values = {value_tokens[0].value: value_tokens[1].value}
            else:
                # list of values
                values = [v.value for v in value_tokens]  # type: ignore
                d["__tokens__"] = [key_token] + list(value_tokens)
        else:
            # single value
            value_token = value_tokens[0]
            # store the original tokens so they can be processed
            # differently for METADATA, VALIDATION, and VALUES
            d["__tokens__"] = [key_token, value_token]
            values = value_token.value

            if self.quoter.is_string(values):
                values = self.clean_string(values)  # type: ignore

        d[key_name] = values

        return d

    def check_composite_tokens(
        self, name: str, tokens: list[Any]
    ) -> tuple[str, list[Any]]:
        """
        Return the key and contents of a KEY..END block
        for PATTERN, POINTS, and PROJECTION
        """
        assert len(tokens) >= 2
        key = tokens[0]

        assert key.value.lower() == name
        assert tokens[-1].value.lower() == "end"

        if len(tokens) == 2:
            body = []  # empty TYPE..END block
        else:
            body = tokens[1:-1]

        body_tokens = []

        for t in body:
            if isinstance(t, dict):
                body_tokens.append(t["__tokens__"])
            else:
                body_tokens.append(t)
        return key, body_tokens

    def process_value_pairs(self, tokens, type_) -> dict:
        """
        Metadata, Values, and Validation blocks can either
        have string pairs or attributes
        Attributes will already be processed
        """
        key, body = self.check_composite_tokens(type_, tokens)
        key_name = self.key_name(key)

        d = CaseInsensitiveOrderedDict(CaseInsensitiveOrderedDict)

        for t in body:
            k = self.clean_string(t[0].value).lower()
            v = self.clean_string(t[1].value)

            if k in d.keys():
                log.warning(
                    "A duplicate key (%s) was found in %s. Only the last value (%s) will be used. ",
                    k,
                    type_,
                    v,
                )

            d[k] = v

        if self.include_position:
            pd = self.create_position_dict(key, body)
            d["__position__"] = pd

        d["__type__"] = key_name

        # return the token as well as the processed dict so the
        # composites function works the same way
        return d

    def connectionoptions(self, tokens):
        """
        Create a dict for the connectionoptions items
        """

        return self.process_value_pairs(tokens, "connectionoptions")

    def metadata(self, tokens):
        """
        Create a dict for the metadata items
        """

        return self.process_value_pairs(tokens, "metadata")

    def values(self, tokens):
        """
        Create a dict for the values items
        """

        return self.process_value_pairs(tokens, "values")

    def config(self, t):
        # process this as a separate rule
        assert len(t) == 3
        key = t[1].value.lower()  # store all subkeys in lowercase
        value = t[2].value
        t[1].value = self.clean_string(key)
        t[2].value = self.clean_string(value)
        return self.attr(t)

    def validation(self, tokens):
        """
        Create a dict for the validation items
        """
        return self.process_value_pairs(tokens, "validation")

    def projection(self, tokens):
        _, body = self.check_composite_tokens("projection", tokens)
        projection_strings = [self.clean_string(v.value) for v in body]

        key_token = tokens[0]
        value_token = tokens[1]  # take the first string as the default token
        value_token = update_token_value(value_token, projection_strings)
        tokens = (key_token, value_token)

        return self.attr(tokens)

    def process_pair_lists(self, key_name, tokens):
        _, body = self.check_composite_tokens(key_name, tokens)
        pairs = [(v[0].value, v[1].value) for v in body]

        key_token = tokens[0]
        value_token = tokens[1][0]  # take the first numeric value pair as the token
        value_token.value = pairs  # set its value to all values
        tokens = (key_token, value_token)

        return self.attr(tokens)

    def points(self, tokens):
        return self.process_pair_lists("points", tokens)

    def pattern(self, tokens):
        return self.process_pair_lists("pattern", tokens)

    # for expressions

    def comparison(self, t):
        assert len(t) == 3
        parts = [str(p.value) for p in t]
        v = " ".join(parts)

        v = f"( {v} )"
        t[0].value = v
        return t[0]

    def and_test(self, t):
        assert len(t) == 2
        t[0].value = f"( {t[0].value} AND {t[1].value} )"
        return t[0]

    def or_test(self, t):
        assert len(t) == 2
        t[0].value = f"( {t[0].value} OR {t[1].value} )"
        return t[0]

    def compare_op(self, t):
        v = t[0]
        return v

    def not_expression(self, t):
        v = t[0]
        v.value = f"NOT {v.value}"
        return v

    def expression(self, t):
        exp = " ".join(
            [str(v.value) for v in t]
        )  # convert to string for boolean expressions e.g. (true)

        if not self.quoter.in_parenthesis(exp):
            t[0].value = f"({exp})"

        return t[0]

    def add(self, t):
        assert len(t) == 2
        t[0].value = f"{t[0].value} + {t[1].value}"
        return t[0]

    def sub(self, t):
        assert len(t) == 2
        t[0].value = f"{t[0].value} - {t[1].value}"
        return t[0]

    def div(self, t):
        assert len(t) == 2
        t[0].value = f"{t[0].value} / {t[1].value}"
        return t[0]

    def mul(self, t):
        assert len(t) == 2
        t[0].value = f"{t[0].value} * {t[1].value}"
        return t[0]

    def power(self, t):
        assert len(t) == 2
        t[0].value = f"{t[0].value} ^ {t[1].value}"
        return t[0]

    def neg(self, t):
        assert len(t) == 1
        t[0].value = f"-{t[0].value}"
        return t[0]

    def runtime_var(self, t):
        v = t[0]
        return v

    def regexp(self, t):
        """
        E.g. regexp(u'/^[0-9]*$/')
        """
        v = t[0]
        return v

    # for functions

    def func_call(self, t):
        """
        For function calls e.g. TEXT (tostring([area],"%.2f"))
        """
        func, params = t
        func_name = func.value
        func.value = f"({func_name}({params}))"
        return func

    def func_params(self, t):
        params = ",".join(str(v.value) for v in t)
        return params

    def attr_bind(self, t):
        assert len(t) == 1
        t = t[0]
        t.value = f"[{t.value}]"
        return t

    def extent(self, t):
        assert len(t) == 4
        return t

    def color(self, t):
        pass

    def value(self, t):
        return t

    # basic types

    def true(self, t):
        v = t[0]
        return update_token_value(v, True)

    def false(self, t):
        v = t[0]
        return update_token_value(v, False)

    def int(self, t):
        v = t[0]
        return update_token_value(v, int(v.value))

    def float(self, t):
        v = t[0]
        return update_token_value(v, float(v.value))

    def bare_string(self, t):
        return t[0]

    def name(self, t):
        v = t[0]
        return v

    def string(self, t):
        v = t[0]
        return v

    def path(self, t):
        return t[0]

    def string_pair(self, t):
        a, b = t
        return [a, b]

    def rgb(self, t):
        r, g, b = t
        return r, g, b

    def attr_bind_pair(self, t):
        assert len(t) == 2
        return t

    def attr_mixed_pair(self, t):
        assert len(t) == 2
        return t

    def colorrange(self, t):
        assert len(t) == 6
        return t

    def hexcolorrange(self, t):
        assert len(t) == 2
        return t

    def hexcolor(self, t):
        t[0].value = self.clean_string(t[0].value).lower()
        return t[0]

    def num_pair(self, t):
        a, b = t
        return a, b

    def int_pair(self, t):
        a, b = t
        return a, b

    def list(self, t):
        # http://www.mapserver.org/mapfile/expressions.html#list-expressions
        v = t[0]
        list_values = ",".join([str(s) for s in t])
        v.value = "{%s}" % list_values
        return v

    def classauto(self, t):
        key_token = t[0]
        key_name = self.key_name(key_token)
        pd = self.create_position_dict(key_token, None)
        d: dict = OrderedDict()
        d["__position__"] = pd
        d[key_name] = None
        return d


class CommentsTransformer(Transformer_InPlace):
    """
    This is an additional transformer that is used to go through
    all the nodes and take any of the custom node.meta.comment properties
    and add these value to the node's corresponding dictionary
    """

    def __init__(self, mapfile_todict):
        self._mapfile_todict = mapfile_todict

    def get_comments(self, meta):
        all_comments = []

        if hasattr(meta, "comments"):
            all_comments += meta.comments

        return all_comments

    def add_metadata_comments(self, d, metadata):
        """
        Any duplicate keys will be replaced with the last duplicate along with comments
        """

        if len(metadata) > 2:
            string_pairs = metadata[1:-1]  # get all metadata pairs
            for sp in string_pairs:
                # get the raw metadata key

                if isinstance(sp.children[0], TOKEN_TYPES):
                    token = sp.children[0]
                    assert token.type == "UNQUOTED_STRING"
                    key = token.value
                else:
                    # quoted string (double or single)
                    token = sp.children[0].children[0]
                    key = token.value

                # clean it to match the dict key
                key = self._mapfile_todict.clean_string(key).lower()
                assert key in d.keys()
                key_comments = self.get_comments(sp.meta)
                d["__comments__"][key] = key_comments

        return d

    @v_args(tree=True)
    def _save_attr_comments(self, tree):
        d = self._mapfile_todict.transform(tree)
        d["__comments__"] = self.get_comments(tree.meta)
        return d

    @v_args(tree=True)
    def _save_composite_comments(self, tree):
        d = self._mapfile_todict.transform(tree)

        # composites such as METADATA will not have had a comments
        # dict created yet
        if "__comments__" not in d:
            d["__comments__"] = OrderedDict()

        comments = self.get_comments(tree.meta)
        if comments:
            d["__comments__"]["__type__"] = comments

        if d["__type__"] == "metadata":
            md = tree.children[0].children
            self.add_metadata_comments(d, md)

        return d

    @v_args(tree=True)
    def _save_projection_comments(self, tree):
        d = self._mapfile_todict.transform(tree)
        comments = self.get_comments(tree.meta)
        if comments:
            d["__comments__"] = comments

        return d

    # below we assign callbacks to process comments for each of the following types
    attr = _save_attr_comments
    config_attr = _save_attr_comments  # handled the same as other attr comments
    projection = _save_projection_comments
    composite = _save_composite_comments


class ConfigfileTransformer(MapfileTransformer):
    def config(self, tree):
        composite_dict = CaseInsensitiveOrderedDict(CaseInsensitiveOrderedDict)
        composite_dict["__type__"] = "config"

        for t in tree:
            key = t.data.lower()

            atts_dict: OrderedDict[str, Any] = OrderedDict()

            if self.include_comments:
                comments_dict = atts_dict["__comments__"] = OrderedDict()

            if self.include_position:
                # self.create_position_dict(key_name, None)
                position_dict = atts_dict["__position__"] = OrderedDict()

            for c in t.children:
                # first remove dicts that are no longer required
                pos = c.pop("__position__")
                c.pop(
                    "__tokens__", None
                )  # tokens are no longer needed now we have the positions
                comments = c.pop("__comments__", None)

                #  simple attribute
                assert len(c.items()) == 1
                att = list(c.items())[0]
                att_key = att[0]
                att_value = att[1]

                if self.include_position:
                    # hoist position details to composite
                    position_dict[att_key] = pos
                if self.include_comments and comments:
                    # hoist comments to composite
                    comments_dict[att_key] = comments

                if att_key in atts_dict.keys():
                    log.warning(
                        "A duplicate key (%s) was found in %s. Only the last value (%s) will be used. ",
                        att_key,
                        key,
                        att_value,
                    )
                atts_dict[att_key] = att_value

            composite_dict[key] = atts_dict

        return composite_dict

    def config_attr(self, tokens) -> dict:
        """
        Process CONFIG file attributes which can be quoted
        or unquoted, otherwise they are identical to other attributes
        """
        return self.attr(tokens)


class MapfileToDict:
    def __init__(
        self,
        include_position=False,
        include_comments=False,
        transformer_class=MapfileTransformer,
        **kwargs,
    ):
        self.include_position = include_position
        self.include_comments = include_comments
        self.transformer_class = transformer_class
        self.kwargs = kwargs

    def transform(self, tree):
        tree = Canonize().transform(tree)

        self.mapfile_transformer = self.transformer_class(
            include_position=self.include_position,
            include_comments=self.include_comments,
            **self.kwargs,
        )

        if self.include_comments:
            comments_transformer = CommentsTransformer(self.mapfile_transformer)
            tree = comments_transformer.transform(tree)

        return self.mapfile_transformer.transform(tree)


class Canonize(Transformer_InPlace):
    @v_args(tree=True)
    def symbolset(self, tree):
        composite_type = Tree("composite_type", [Token("symbolset", "symbolset")])

        tree.data = "composite"
        tree.children.insert(0, composite_type)
        return tree


def calculate_depth(iterable):
    """
    A helper function to get the max length + 1 in a list of lists
    """
    return (
        isinstance(iterable, (tuple, list)) and max(map(calculate_depth, iterable)) + 1
    )



# =================================================================
#
# Authors: Seth Girvin
#
# Copyright (c) 2023 Seth Girvin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from __future__ import annotations
import logging
from typing import Any


log = logging.getLogger("mappyfile")


class Quoter:
    """
    A class to handle adding and standardising quotes around strings
    """

    def __init__(self, quote: str = '"'):
        assert quote in ("'", '"')

        self.quote = quote

        if self.quote == "'":
            self.altquote = '"'
        else:
            self.altquote = "'"

    def add_quotes(self, val: str) -> str:
        return self._add_quotes(val, self.quote)

    def add_altquotes(self, val: str) -> str:
        return self._add_quotes(val, self.altquote)

    def _add_quotes(self, val: str, quote: str) -> str:
        return f"{quote}{val}{quote}"

    def in_quotes(self, val: str) -> bool:
        return self._in_quotes(val, self.quote) or self._in_quotes(val, self.altquote)

    def _in_quotes(self, val: str, char: str):
        return val.startswith(char) and val.endswith(char)

    def escape_quotes(self, val: Any) -> Any:
        """
        Escape any quotes in a value
        """
        if self.is_string(val) and self._in_quotes(val, self.quote):
            # make sure any previously escaped quotes are not re-escaped
            middle = self.remove_quotes(val).replace("\\" + self.quote, self.quote)
            middle = middle.replace(self.quote, "\\" + self.quote)
            val = self.add_quotes(middle)

        return val

    def is_string(self, val: Any) -> bool:
        return isinstance(val, str)

    def remove_quotes(self, val: Any) -> Any:
        if isinstance(val, list):
            return list(map(self.remove_quotes, val))

        if not self.is_string(val):
            return val

        if self.in_quotes(val):
            return val[1:-1]

        return val

    def in_brackets(self, val: str) -> bool:
        val = val.strip()
        return val.startswith("[") and val.endswith("]")

    def in_parenthesis(self, val: str) -> bool:
        val = val.strip()
        return val.startswith("(") and val.endswith(")")

    def in_braces(self, val: str) -> bool:
        val = val.strip()
        return val.startswith("{") and val.endswith("}")

    def in_slashes(self, val: str) -> bool:
        val = val.strip()
        return self._in_quotes(val, "/")

    def standardise_quotes(self, val: str) -> str:
        """
        Change the quotes used to wrap a value to the pprint default
        E.g. "val" to 'val' or 'val' to "val"
        """
        if self._in_quotes(val, self.altquote):
            middle = self.remove_quotes(val)
            val = self.add_quotes(middle)

        return self.escape_quotes(val)



# =================================================================
#
# Authors: Seth Girvin
#
# Copyright (c) 2021 Seth Girvin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from __future__ import annotations
import codecs
import warnings
import functools
from mappyfile.ordereddict import DefaultOrderedDict
from mappyfile.parser import Parser
from mappyfile.transformer import (
    MapfileToDict,
    MapfileTransformer,
    ConfigfileTransformer,
)
from mappyfile.pprint import PrettyPrinter
from mappyfile.validator import Validator
from typing import IO, Any, Type


def deprecated(func):
    """
    From https://stackoverflow.com/questions/2536307/how-do-i-deprecate-python-functions/30253848#30253848
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter("always", DeprecationWarning)  # turn off filter
        warnings.warn(
            f"Call to deprecated function {func.__name__}.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func


def _transform(
    ast,
    include_comments: bool = False,
    include_position: bool = False,
    **kwargs,
) -> dict:
    transformer_class: Type[Any]
    if "transformer_class" not in kwargs:
        if ast.data and ast.data == "config":
            transformer_class = ConfigfileTransformer
        else:
            transformer_class = MapfileTransformer
    else:
        # a transformer_class was set as an argument
        transformer_class = kwargs.pop("transformer_class")

    m = MapfileToDict(
        include_position=include_position,
        include_comments=include_comments,
        transformer_class=transformer_class,
        **kwargs,
    )
    d = m.transform(ast)
    return d


# pylint: disable=redefined-builtin
def open(
    fn: str,
    expand_includes: bool = True,
    include_comments: bool = False,
    include_position: bool = False,
    **kwargs,
) -> dict:
    """
    Load a Mapfile from the supplied filename into a Python dictionary.

    Parameters
    ----------

    fn: string
        The path to the Mapfile, or partial Mapfile
    expand_includes: boolean
        Load any ``INCLUDE`` files in the MapFile
    include_comments: boolean
         Include or discard comment strings from the Mapfile - *experimental*
    include_position: boolean
         Include the position of the Mapfile tokens in the output

    Returns
    -------

    dict
        A Python dictionary representing the Mapfile in the mappyfile format

    Example
    -------

    To open a Mapfile from a filename and return it as a dictionary object::

        d = mappyfile.open('mymap.map')

    Notes
    -----

    Partial Mapfiles can also be opened, for example a file containing a ``LAYER`` object.

    """

    p = Parser(
        expand_includes=expand_includes, include_comments=include_comments, **kwargs
    )
    ast = p.parse_file(fn)
    return _transform(ast, include_comments, include_position, **kwargs)


def load(
    fp: IO[str],
    expand_includes: bool = True,
    include_position: bool = False,
    include_comments: bool = False,
    **kwargs,
) -> dict:
    """
    Load a Mapfile from an open file or file-like object.

    Parameters
    ----------

    fp: file
        A file-like object - as with all Mapfiles this should be encoded in "utf-8"
    expand_includes: boolean
        Load any ``INCLUDE`` files in the MapFile
    include_comments: boolean
         Include or discard comment strings from the Mapfile - *experimental*
    include_position: boolean
         Include the position of the Mapfile tokens in the output

    Returns
    -------

    dict
        A Python dictionary representing the Mapfile in the mappyfile format

    Example
    -------

    To open a Mapfile from a file and return it as a dictionary object::

        with open('mymap.map') as fp:
            d = mappyfile.load(fp)

    Notes
    -----

    Partial Mapfiles can also be opened, for example a file containing a ``LAYER`` object.
    """
    p = Parser(
        expand_includes=expand_includes, include_comments=include_comments, **kwargs
    )
    ast = p.load(fp)
    return _transform(ast, include_comments, include_position, **kwargs)


def loads(
    s: str,
    expand_includes: bool = True,
    include_position: bool = False,
    include_comments: bool = False,
    **kwargs,
) -> dict:
    """
    Load a Mapfile from a string

    Parameters
    ----------

    s: string
        The Mapfile, or partial Mapfile, text
    expand_includes: boolean
        Load any ``INCLUDE`` files in the MapFile
    include_comments: boolean
         Include or discard comment strings from the Mapfile - *experimental*
    include_position: boolean
         Include the position of the Mapfile tokens in the output

    Returns
    -------

    dict
        A Python dictionary representing the Mapfile in the mappyfile format

    Example
    -------

    To open a Mapfile from a string and return it as a dictionary object::

        s = '''MAP NAME "TEST" END'''

        d = mappyfile.loads(s)
        assert d["name"] == "TEST"

    """
    p = Parser(
        expand_includes=expand_includes, include_comments=include_comments, **kwargs
    )
    ast = p.parse(s)
    return _transform(ast, include_comments, include_position, **kwargs)


# pylint: disable=too-many-arguments
def dump(
    d: dict,
    fp: IO[str],
    indent: int = 4,
    spacer: str = " ",
    quote: str = '"',
    newlinechar: str = "\n",
    end_comment: bool = False,
    align_values: bool = False,
    separate_complex_types: bool = False,
):
    """
    Write d (the Mapfile dictionary) as a formatted stream to fp

    Parameters
    ----------

    d: dict
        A Python dictionary based on the the mappyfile schema
    fp: file
        A file-like object
    indent: int
        The number of ``spacer`` characters to indent structures in the Mapfile
    spacer: string
        The character to use for indenting structures in the Mapfile. Typically
        spaces or tab characters (``\\t``)
    quote: string
        The quote character to use in the Mapfile (double or single quotes)
    newlinechar: string
        The character used to insert newlines in the Mapfile
    end_comment: bool
        Add a comment with the block type at each closing END
        statement e.g. END # MAP
    align_values: bool
        Aligns the values in the same column for better readability. The column is
        multiple of indent and determined by the longest key
    separate_complex_types: bool
        Groups composites (complex mapserver definitions with "END") together at the end.
        Keeps the given order except that all simple key-value pairs appear before composites.

    Example
    -------

    To open a Mapfile from a string, and then dump it back out to an open file,
    using 2 spaces for indentation, and single-quotes for properties::

        s = '''MAP NAME "TEST" END'''

        d = mappyfile.loads(s)
        with open(fn, "w") as f:
            mappyfile.dump(d, f, indent=2, quote="'")

    """
    map_string = _pprint(
        d,
        indent,
        spacer,
        quote,
        newlinechar,
        end_comment,
        align_values,
        separate_complex_types,
    )
    fp.write(map_string)


# pylint: disable=too-many-arguments
def save(
    d: dict,
    output_file: str,
    indent: int = 4,
    spacer: str = " ",
    quote: str = '"',
    newlinechar: str = "\n",
    end_comment: bool = False,
    align_values: bool = False,
    separate_complex_types: bool = False,
) -> str:
    """
    Write a dictionary to an output Mapfile on disk

    Parameters
    ----------

    d: dict
        A Python dictionary based on the the mappyfile schema
    output_file: string
        The output filename
    indent: int
        The number of ``spacer`` characters to indent structures in the Mapfile
    spacer: string
        The character to use for indenting structures in the Mapfile. Typically
        spaces or tab characters (``\\t``)
    quote: string
        The quote character to use in the Mapfile (double or single quotes)
    newlinechar: string
        The character used to insert newlines in the Mapfile
    end_comment: bool
        Add a comment with the block type at each closing END
        statement e.g. END # MAP
    align_values: bool
        Aligns the values in the same column for better readability. The column is
        multiple of indent and determined by the longest key.
    separate_complex_types: bool
        Groups composites (complex mapserver definitions with "END") together at the end.
        Keeps the given order except that all simple key-value pairs appear before composites.

    Returns
    -------

    string
          The output_file passed into the function

    Example
    -------

    To open a Mapfile from a string, and then save it to a file::

        s = '''MAP NAME "TEST" END'''

        d = mappyfile.loads(s)
        fn = "C:/Data/mymap.map"
        mappyfile.save(d, fn)
    """
    map_string = _pprint(
        d,
        indent,
        spacer,
        quote,
        newlinechar,
        end_comment,
        align_values,
        separate_complex_types,
    )
    _save(output_file, map_string)
    return output_file


# pylint: disable=too-many-arguments
def dumps(
    d: dict,
    indent: int = 4,
    spacer: str = " ",
    quote: str = '"',
    newlinechar: str = "\n",
    end_comment: bool = False,
    align_values: bool = False,
    separate_complex_types: bool = False,
    **kwargs,
) -> str:
    """
    Output a Mapfile dictionary as a string

    Parameters
    ----------

    d: dict
        A Python dictionary based on the the mappyfile schema
    indent: int
        The number of ``spacer`` characters to indent structures in the Mapfile
    spacer: string
        The character to use for indenting structures in the Mapfile. Typically
        spaces or tab characters (``\\t``)
    quote: string
        The quote character to use in the Mapfile (double or single quotes)
    newlinechar: string
        The character used to insert newlines in the Mapfile
    end_comment: bool
        Add a comment with the block type at each closing END
        statement e.g. END # MAP
    align_values: bool
        Aligns the values in the same column for better readability. The column is
        multiple of indent and determined by the longest key
    separate_complex_types: bool
        Groups composites (complex mapserver definitions with "END") together at the end.
        Keeps the given order except that all simple key-value pairs appear before composites.

    Returns
    -------

    string
          The Mapfile as a string

    Example
    -------

    To open a Mapfile from a string, and then print it back out
    as a string using tabs::

        s = '''MAP NAME "TEST" END'''

        d = mappyfile.loads(s)
        print(mappyfile.dumps(d, indent=1, spacer="\\t"))
    """
    return _pprint(
        d,
        indent,
        spacer,
        quote,
        newlinechar,
        end_comment,
        align_values,
        separate_complex_types,
        **kwargs,
    )


def validate(d: dict, version: float | None = None, schema_name: str = "map") -> list:
    """
    Validate a mappyfile dictionary by using the Mapfile schema.
    An optional version number can be used to specify a specific
    a Mapfile is valid for a specific MapServer version.

    Parameters
    ----------

     d: dict
        A Python dictionary based on the the mappyfile schema
    version: float
        The MapServer version number used to validate the Mapfile

    Returns
    -------

    list
        A list containing validation errors

    """
    v = Validator()
    return v.validate(d, version=version, schema_name=schema_name)


def _save(output_file: str, string: str) -> None:
    with codecs.open(output_file, "w", encoding="utf-8") as f:
        f.write(string)


def _pprint(
    d: dict,
    indent: int,
    spacer: str,
    quote: str,
    newlinechar: str,
    end_comment: bool,
    align_values: bool,
    separate_complex_types: bool,
    **kwargs,
) -> str:
    pp = PrettyPrinter(
        indent=indent,
        spacer=spacer,
        quote=quote,
        newlinechar=newlinechar,
        end_comment=end_comment,
        align_values=align_values,
        separate_complex_types=separate_complex_types,
        **kwargs,
    )
    return pp.pprint(d)


def create(type: str, version=None, add_defaults=False) -> dict:
    """
    Create a new mappyfile object. To add MapServer defaults (if any)
    set ``add_defaults`` to ``True``.

    Parameters
    ----------

    s: type
        The mappyfile type to be stored in the __type__ property

    Returns
    -------

    dict
        A Python dictionary representing the Mapfile object in the mappyfile format
    """

    # get the schema for this type

    v = Validator()
    try:
        schema = v.get_versioned_schema(version=version, schema_name=type)
    except IOError as exc:
        raise SyntaxError("The mappyfile type '{type}' does not exist!") from exc

    d = DefaultOrderedDict()
    d["__type__"] = type

    if add_defaults is True:
        properties = sorted(schema["properties"].items())
        for key, value in properties:
            if "default" in value:
                d[key] = value["default"]

    return d



# =================================================================
#
# Authors: Seth Girvin
#
# Copyright (c) 2020 Seth Girvin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging
import importlib.metadata
from logging import NullHandler
import sys
from types import ModuleType

# allow high-level functions to be accessed directly from the mappyfile module
# pylint: disable=redefined-builtin
from mappyfile.utils import (
    open,
    create,
    load,
    loads,
    dumps,
    dump,
    save,
    validate,
)
from mappyfile.dictutils import (
    find,
    findall,
    findkey,
    findunique,
    update,
    dict_move_to_end,
)

__version__ = "1.1.0"

__all__ = [
    "open",
    "load",
    "loads",
    "find",
    "findall",
    "findunique",
    "dumps",
    "dump",
    "save",
    "findkey",
    "update",
    "validate",
    "create",
    "dict_move_to_end",
]


module = ModuleType("mappyfile.plugins")
sys.modules["mappyfile.plugins"] = module

if sys.version_info >= (3, 10):
    plugins = importlib.metadata.entry_points(group="mappyfile.plugins")
else:
    plugins = importlib.metadata.entry_points().get("mappyfile.plugins", [])

for plugin in plugins:
    setattr(module, plugin.name, plugin.load())

# Set default logging handler to avoid "No handler found" warnings.
logging.getLogger("mappyfile").addHandler(NullHandler())



# =================================================================
#
# Authors: Seth Girvin
#
# Copyright (c) 2020 Seth Girvin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import sys
import os
import codecs
import json
import glob
import logging
import click
import mappyfile
from mappyfile.validator import Validator


def get_mapfiles(mapfiles):
    all_mapfiles = [
        mf for sublist in mapfiles for mf in glob.glob(sublist) if not os.path.isdir(mf)
    ]
    return all_mapfiles


def configure_logging(verbosity: int):
    """
    Configure logging level

    Parameters
    ----------
    verbosity : int
        The number of `-v` options from the command line.
    Returns
    -------
    None
    """
    log_level = max(10, 30 - 10 * verbosity)
    logging.basicConfig(stream=sys.stderr, level=log_level)


logger = logging.getLogger(__name__)


# The CLI command group.
@click.group(help="Command line interface for the mappyfile package")
@click.option("--verbose", "-v", count=True, help="Increase verbosity")
@click.option("--quiet", "-q", count=True, help="Decrease verbosity")
@click.version_option(version=mappyfile.__version__, message="%(version)s")
@click.pass_context
def main(ctx, verbose, quiet):
    """
    Execute the main mappyfile command
    """
    verbosity = verbose - quiet
    configure_logging(verbosity)
    ctx.obj = {}
    ctx.obj["verbosity"] = verbosity


# pylint: disable=redefined-builtin
@main.command(short_help="Format a Mapfile")
@click.argument("input-mapfile", nargs=1, type=click.Path(exists=True))
@click.argument("output-mapfile", nargs=1, type=click.Path())
@click.option(
    "--indent",
    default=4,
    show_default=True,
    help="The number of spacer characters to indent structures in the Mapfile",
)  # noqa
@click.option(
    "--spacer",
    default=" ",
    help="The character to use for indenting structures in the Mapfile",
)
@click.option(
    "--quote",
    default='"',
    help='The quote character to use in the Mapfile (double or single quotes). Ensure these are escaped e.g. \\" or \\\' [default: \\"]',
)  # noqa
@click.option(
    "--newlinechar",
    default="\n",
    help="The character used to insert newlines in the Mapfile [default: \\n]",
)
@click.option(
    "--expand/--no-expand",
    default=True,
    show_default=True,
    help="Expand any INCLUDE directives found in the Mapfile",
)  # noqa
@click.option(
    "--comments/--no-comments",
    default=False,
    show_default=True,
    help="Keep Mapfile comments in the output (experimental)",
)  # noqa
@click.pass_context
# pylint: disable=too-many-arguments
def format(
    _,
    input_mapfile,
    output_mapfile,
    indent,
    spacer,
    quote,
    newlinechar,
    expand,
    comments,
):
    """
    Format a the input-mapfile and save as output-mapfile. Note output-mapfile will be
    overwritten if it already exists.

    Example of formatting a single Mapfile:

        mappyfile format C:/Temp/valid.map C:/Temp/valid_formatted.map

    Example of formatting a single Mapfile with single quotes and tabs for indentation:

        mappyfile format C:/Temp/valid.map C:/Temp/valid_formatted.map --quote=\\' --indent=1 --spacer=\t

    Example of formatting a single Mapfile without expanding includes, but including comments:

        mappyfile format C:/Temp/valid.map C:/Temp/valid_formatted.map --no-expand --comments
    """

    quote = codecs.decode(quote, "unicode_escape")  # ensure \t is handled as a tab
    spacer = codecs.decode(spacer, "unicode_escape")  # ensure \t is handled as a tab
    newlinechar = codecs.decode(
        newlinechar, "unicode_escape"
    )  # ensure \n is handled as a newline

    d = mappyfile.open(
        input_mapfile,
        expand_includes=expand,
        include_comments=comments,
        include_position=True,
    )
    mappyfile.save(
        d,
        output_mapfile,
        indent=indent,
        spacer=spacer,
        quote=quote,
        newlinechar=newlinechar,
    )
    sys.exit(0)


@main.command(short_help="Validate Mapfile(s) against a schema")
@click.argument("mapfiles", nargs=-1, type=click.Path())
@click.option(
    "--expand/--no-expand",
    default=True,
    show_default=True,
    help="Expand any INCLUDE directives found in the Mapfile",
)  # noqa
@click.option(
    "--version",
    default=8.6,
    show_default=True,
    help="The MapServer version number used to validate the Mapfile",
)  # noqa
@click.pass_context
def validate(_, mapfiles, expand, version):
    """
    Validate Mapfile(s) against the Mapfile schema

    The MAPFILES argument is a list of paths, either to individual Mapfiles, or a folders containing Mapfiles.
    Wildcards are supported (natively on Linux, and up to one level deep on Windows).
    Validation errors are reported to the console. The program returns the error count - this will be 0 if no
    validation errors are encountered.

    Example of validating a single Mapfile:

        mappyfile validate C:/Temp/valid.map

    Example of validating two folders containing Mapfiles, without expanding INCLUDES:

        mappyfile validate C:/Temp/*.map D:/GitHub/mappyfile/tests/mapfiles/*.map --no-expand
    """

    all_mapfiles = get_mapfiles(mapfiles)

    if len(all_mapfiles) == 0:
        click.echo(f"No Mapfiles found at the following paths: {','.join(mapfiles)}")
        return

    validation_count = 0
    errors = 0

    for fn in all_mapfiles:
        fn = click.format_filename(fn)
        try:
            d = mappyfile.open(fn, expand_includes=expand, include_position=True)
        except Exception as ex:
            logger.exception(ex)
            click.echo(f"{fn} failed to parse successfully")
            continue

        schema_name = d.get("__type__", "map")
        validation_messages = mappyfile.validate(d, version, schema_name)
        if validation_messages:
            for v in validation_messages:
                v["fn"] = fn
                msg = (
                    "{fn} (Line: {line} Column: {column}) {message} - {error}"
                ).format(
                    fn=v.get("fn", "<unknown>"),
                    line=v.get("line", "?"),
                    column=v.get("column", "?"),
                    message=v.get("message", ""),
                    error=v.get("error", ""),
                )
                click.echo(msg)
                errors += 1
        else:
            click.echo(f"{fn} validated successfully")
            validation_count += 1

    click.echo(
        f"{len(all_mapfiles)} file(s) validated ({validation_count} successfully)"
    )
    sys.exit(errors)


@main.command(short_help="Export a Mapfile Schema")
@click.argument("output_file", nargs=1, type=click.Path())
@click.option(
    "--version",
    type=click.FLOAT,
    help="The MapServer version number used to validate the Mapfile",
)  # noqa
@click.pass_context
def schema(_, output_file, version=None):
    """
    Save the Mapfile schema to a file. Set the version parameter to output a specific version.
    Note output-file will be overwritten if it already exists.

    Examples:

        mappyfile schema C:/Temp/mapfile-schema.json

    Example of a specific version:

        mappyfile schema C:/Temp/mapfile-schema-7-6.json --version=7.6
    """
    validator = Validator()
    jsn = validator.get_versioned_schema(version)

    with codecs.open(output_file, "w", encoding="utf-8") as f:
        f.write(json.dumps(jsn, sort_keys=True, indent=4))

    sys.exit(0)



# =================================================================
#
# Authors: Seth Girvin
#
# Copyright (c) 2020 Seth Girvin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from __future__ import annotations
import logging
import numbers
from mappyfile.tokens import (
    COMPOSITE_NAMES,
    SINGLETON_COMPOSITE_NAMES,
    CONFIG_COMPOSITE_NAMES,
    REPEATED_KEYS,
    COMPLEX_TYPES,
    OBJECT_LIST_KEYS,
)
from mappyfile.validator import Validator
from mappyfile.quoter import Quoter
from mappyfile import dictutils
from typing import Any


log = logging.getLogger("mappyfile")


# pylint: disable=too-many-arguments
class PrettyPrinter:
    def __init__(
        self,
        indent: int = 4,
        spacer: str = " ",
        quote: str = '"',
        newlinechar: str = "\n",
        end_comment: bool = False,
        align_values: bool = False,
        separate_complex_types: bool = False,
    ):
        """
        Option use "\t" for spacer with an indent of 1
        """

        assert quote in ("'", '"')

        self.indent = indent
        self.spacer = spacer * self.indent
        self.quoter = Quoter(quote)
        self.newlinechar = newlinechar
        self.end_comment = end_comment
        self.end = "END"
        self.validator = Validator()
        self.align_values = align_values
        self.separate_complex_types = separate_complex_types

    def __is_metadata(self, key: str) -> bool:
        """
        Check to see if the property is hidden metadata
        e.g. "__type__", "__comments__", "__position__"
        """
        if key.startswith("__") and key.endswith("__"):
            return True

        return False

    def compute_aligned_max_indent(self, max_key_length: int) -> int:
        """
        Computes the indentation as a multiple of self.indent for aligning
        values at the same column based on the maximum key length.
        Example:
        key         value1
        longkey     value2
        longestkey  value3 <-- column at 12, indent of 4, determined by "longestkey"
        """
        indent = max(1, self.indent)
        return int((int(max_key_length / indent) + 1) * indent)

    def compute_max_key_length(self, composite: dict) -> int:
        """
        Computes the maximum length of all keys (non-recursive) in the passed
        composite.
        """
        length = 0

        ignore_list = (
            "metadata",
            "validation",
            "values",
            "connectionoptions",
            "pattern",
            "projection",
            "points",
            "config",
        )

        for attr, value in composite.items():
            attr_length = len(attr)
            if (
                not self.__is_metadata(attr)
                and attr not in ignore_list
                and not self.is_hidden_container(attr, value)
                and not self.is_composite(value)
            ):
                length = max(length, attr_length)

        return length

    def separate_complex(self, composite: dict, level: int) -> None:
        if not self.separate_complex_types:
            return
        for key in list(composite.keys()):
            if self.is_complex_type(composite, key, level):
                dictutils.dict_move_to_end(composite, key)

    def whitespace(self, level: int, indent: int) -> str:
        return self.spacer * (level + indent)

    def add_start_line(self, key: str, level: int) -> str:
        return self.whitespace(level, 1) + key.upper()

    def add_end_line(self, level: int, indent: int, key: str) -> str:
        end_line = self.whitespace(level, indent) + self.end
        if self.end_comment:
            end_line = f"{end_line} # {key.upper()}"
        return end_line

    def __format_line(
        self, spacer: str, key: str, value: Any, aligned_max_indent: int = 0
    ) -> str:
        if (aligned_max_indent is None) or (aligned_max_indent == 0):
            aligned_max_indent = len(key) + 1
        if value == "":
            indent = ""  # for keywords with no values, do not add a space
        else:
            indent = " " * (aligned_max_indent - len(key))

        tmpl = "{spacer}{key}{indent}{value}"
        d = {"spacer": spacer, "key": key, "value": value, "indent": indent}
        return tmpl.format(**d)

    def process_key_dict(
        self, key: str, d: dict, level: int, quote_key: bool = True
    ) -> list[str]:
        """
        Process key value dicts e.g. METADATA "key" "value"
        """

        # add any composite level comments
        comments = d.get("__comments__", {})
        lines: list[str] = []
        self._add_type_comment(level, comments, lines)

        lines += [self.add_start_line(key, level)]
        lines += self.process_dict(d, level, comments, quote_key)
        lines.append(self.add_end_line(level, 1, key))

        return lines

    def process_dict(
        self, d: dict, level: int, comments: dict, quote_key: bool = True
    ) -> list[str]:
        """
        Process keys and values within a block
        """
        lines = []

        aligned_max_indent = 0
        if self.align_values:
            max_key_length = self.compute_max_key_length(d) + 2  # add length of quotes
            aligned_max_indent = self.compute_aligned_max_indent(max_key_length)

        for k, v in d.items():
            if not self.__is_metadata(k):
                if quote_key:
                    qk = self.quoter.add_quotes(k)
                else:
                    qk = k
                qv = self.quoter.add_quotes(v)
                line = self.__format_line(
                    self.whitespace(level, 2), qk, qv, aligned_max_indent
                )
                line += self.process_attribute_comment(comments, k)
                lines.append(line)

        return lines

    def process_config_dict(self, d: dict, level: int) -> list[str]:
        """
        Process the CONFIG block
        """
        lines = []
        for k, v in d.items():
            cfg_val = self.quoter.add_quotes(k.upper())
            k = f"CONFIG {cfg_val}"
            v = self.quoter.add_quotes(v)
            lines.append(self.__format_line(self.whitespace(level, 1), k, v))
        return lines

    def process_repeated_list(
        self, key: str, lst: list[str], level: int, aligned_max_indent: int = 1
    ) -> list[str]:
        """
        Process blocks of repeated keys e.g. FORMATOPTION
        """
        lines = []

        for v in lst:
            k = key.upper()
            v = self.quoter.add_quotes(v)
            lines.append(
                self.__format_line(self.whitespace(level, 1), k, v, aligned_max_indent)
            )

        return lines

    def process_projection(
        self, key, lst: str | list[str], level: int, projection_comments: str
    ) -> list[str]:
        lines = [self.add_start_line(key, level)]

        whitespace = self.whitespace(level, 2)

        if projection_comments:
            lines.append(f"{whitespace}{projection_comments.strip()}")

        if self.quoter.is_string(lst):
            val = self.quoter.add_quotes(str(lst))
            # the value has been manually set to a single string projection
            lines.append(f"{whitespace}{val}")
        elif len(lst) == 1 and lst[0].upper() == "AUTO":
            lines.append(f"{whitespace}AUTO")
        else:
            for v in lst:
                v = self.quoter.add_quotes(v)
                lines.append(f"{whitespace}{v}")

        lines.append(self.add_end_line(level, 1, key))
        return lines

    def format_pair_list(self, key: str, pair_list: list[Any], level: int) -> list[str]:
        """
        Process lists of pairs (e.g. PATTERN block)
        """

        lines = [self.add_start_line(key, level)]

        list_spacer = self.spacer * (level + 2)
        pairs = [f"{list_spacer}{p[0]} {p[1]}" for p in pair_list]
        lines += pairs

        lines.append(self.add_end_line(level, 1, key))

        return lines

    def format_repeated_pair_list(
        self, key: str, root_list: list[Any], level: int
    ) -> list[str]:
        """
        Process (possibly) repeated lists of pairs e.g. POINTs blocks
        """

        lines = []

        def depth(iterable):
            return isinstance(iterable, (tuple, list)) and max(map(depth, iterable)) + 1

        if depth(root_list) == 2:
            # single set of points only
            root_list = [root_list]

        for pair_list in root_list:
            lines += self.format_pair_list(key, pair_list, level)

        return lines

    def is_composite(self, val: Any) -> bool:
        if isinstance(val, dict) and "__type__" in val:
            return True
        return False

    def is_complex_type(self, composite: dict, key: str, level: int) -> bool:
        # symbol needs special treatment
        if key == "symbol" and level > 0:
            return False
        return (
            key in COMPLEX_TYPES
            or self.is_composite(key)
            or self.is_hidden_container(key, composite[key])
        )

    def is_hidden_container(self, key: str, val: Any) -> bool:
        """
        The key is not one of the Mapfile keywords, and its
        values are a list
        """

        if key in OBJECT_LIST_KEYS and isinstance(val, list):
            return True
        return False

    def pprint(self, composites: dict | list[dict]) -> str:
        """
        Print out a nicely indented Mapfile
        """

        # if only a single composite is used then cast to list
        # and allow for multiple root composites

        if composites and not isinstance(composites, list):
            composites = [composites]

        lines = []

        for composite in composites:
            type_ = composite["__type__"]
            if type_ in (
                "metadata",
                "validation",
                "connectionoptions",
                "env",
                "maps",
                "plugins",
            ):
                # types are being parsed directly, and not as an attr of a parent
                lines += self.process_key_dict(type_, composite, level=0)
            else:
                lines += self._format(composite)

        result = str(self.newlinechar.join(lines))
        return result

    def get_attribute_properties(self, type_: str, attr: str) -> dict:
        jsn_schema = self.validator.get_expanded_schema(type_)
        props = jsn_schema["properties"]

        # check if a value needs to be quoted or not, by referring to the JSON schema

        try:
            attr_props = props[attr]
        except KeyError as ex:
            log.error(
                "The key '%s' was not found in the JSON schema for '%s'", attr, type_
            )
            log.error(ex)
            return {}

        return attr_props

    def is_expression(self, option):
        return "description" in option and (option["description"] == "expression")

    def check_options_list(self, options_list, value):
        for option in options_list:
            if ("enum" in option and value.lower() in option["enum"]) or (
                "const" in option and value.lower() in option["const"]
            ):
                if value.lower() == "end":
                    # in GEOTRANSFORM "end" is an attribute value
                    return self.quoter.add_quotes(value)
                return value.upper()

            if self.is_expression(option):
                if value.endswith("'i") or value.endswith('"i'):
                    return value

        if self.quoter.in_slashes(value):
            return value

        return self.quoter.add_quotes(value)

    def format_value(self, attr: str, attr_props, value: Any) -> Any:
        """
        TODO - refactor and add more specific tests (particularly for expressions)
        """
        if isinstance(value, bool):
            return str(value).upper()

        if any(i in ["enum"] for i in attr_props):
            if isinstance(value, dict) and not value:
                raise ValueError(
                    f"The property {attr} has an empty dictionary as a value"
                )

            if not isinstance(value, numbers.Number):
                if attr == "compop":
                    return self.quoter.add_quotes(str(value))
                return str(value).upper()  # value is from a set list, no need for quote

            return value

        if "type" in attr_props and attr_props["type"] == "null":
            # for keywords without a value return an empty string
            # currently only CLASSAUTO is the only keyword without a value
            assert attr.lower() == "classauto"
            return ""

        if (
            "type" in attr_props and attr_props["type"] == "string"
        ):  # and "enum" not in attr_props
            # check schemas for expressions and handle accordingly
            if self.is_expression(attr_props) and self.quoter.in_slashes(value):
                return value
            if self.is_expression(attr_props) and (
                value.endswith("'i") or value.endswith('"i')
            ):
                # for case insensitive regex
                return value

            return self.quoter.add_quotes(value)

        # expressions can be one of a string or an expression in brackets
        if any(
            i in ["oneOf", "anyOf"] for i in attr_props
        ):  # and check that type string is in list
            if "oneOf" in attr_props:
                options_list = attr_props["oneOf"]
            else:
                options_list = attr_props["anyOf"]
            if self.quoter.is_string(value):
                if self.quoter.in_parenthesis(value):
                    pass
                elif attr == "expression" and self.quoter.in_braces(value):
                    # don't add quotes to list expressions such as {val1, val2}
                    pass
                elif attr != "text" and self.quoter.in_brackets(value):
                    # TEXT expressions are often "[field1]-[field2]" so need to leave quotes for these
                    pass
                elif value.startswith("NOT ") and self.quoter.in_parenthesis(value[4:]):
                    value = f"NOT {value[4:]}"
                else:
                    value = self.check_options_list(options_list, value)

        if isinstance(value, list):
            new_values = []

            for v in value:
                if not isinstance(v, numbers.Number) and attr not in [
                    "offset",
                    "polaroffset",
                ]:
                    # don't add quotes to list of attributes for offset / polaroffset
                    v = self.quoter.add_quotes(v)
                new_values.append(v)

            value = " ".join(list(map(str, new_values)))
        else:
            value = self.quoter.escape_quotes(value)

        return value

    # pylint: disable=too-many-arguments
    def process_attribute(
        self, type_: str, attr: str, value: Any, level: int, aligned_max_indent: int = 1
    ) -> str:
        """
        Process one of the main composite types (see the type_ value)
        """

        attr_props = self.get_attribute_properties(type_, attr)
        value = self.format_value(attr, attr_props, value)
        line = self.__format_line(
            self.whitespace(level, 1), attr.upper(), value, aligned_max_indent
        )
        return line

    def format_comment(self, spacer: str, value: str) -> str:
        return f"{spacer}{value}"

    def process_composite_comment(self, level: int, comments: dict, key: str) -> str:
        """
        Process comments for composites such as MAP, LAYER etc.
        """
        if key not in comments:
            comment = ""
        else:
            value = comments[key]
            spacer = self.whitespace(level, 0)

            if isinstance(value, list):
                comment_list = [self.format_comment(spacer, v) for v in value]
                comment = self.newlinechar.join(comment_list)
            else:
                comment = self.format_comment(spacer, value)

        return comment

    def process_attribute_comment(self, comments: dict, key: str) -> str:
        if key not in comments:
            comment = ""
        else:
            value = comments[key]
            spacer = " "

            # for multiple comments associated with an attribute
            # simply join them together as a single string
            if isinstance(value, list):
                value = " ".join(value)

            comment = self.format_comment(spacer, value)

        return comment

    def _add_type_comment(self, level: int, comments: dict, lines: list[str]) -> None:
        comment = self.process_composite_comment(level, comments, "__type__")

        if comment:
            lines.append(str(comment))

    def _format(self, composite: dict, level: int = 0) -> list[str]:
        lines: list[str] = []
        type_ = ""

        # get any comments associated with the composite
        comments = composite.get("__comments__", {})

        if isinstance(composite, dict) and "__type__" in composite:
            type_ = composite["__type__"]
            assert type_ in COMPOSITE_NAMES.union(SINGLETON_COMPOSITE_NAMES).union(
                CONFIG_COMPOSITE_NAMES
            )
            is_hidden = False
            self._add_type_comment(level, comments, lines)
            s = self.whitespace(level, 0) + type_.upper()
            lines.append(s)

        aligned_max_indent = 0
        if self.align_values:
            max_key_length = self.compute_max_key_length(composite)
            aligned_max_indent = self.compute_aligned_max_indent(max_key_length)

        self.separate_complex(composite, level)

        for attr, value in composite.items():
            if self.__is_metadata(attr):
                # skip hidden attributes
                continue

            if self.is_hidden_container(attr, value):
                # now recursively print all the items in the container
                for v in value:
                    lines += self._format(v, level + 1)
            elif attr == "pattern":
                lines += self.format_pair_list(attr, value, level)
            elif attr in (
                "metadata",
                "validation",
                "values",
                "connectionoptions",
            ):
                # metadata and values are also composites
                # but will be processed here
                lines += self.process_key_dict(attr, value, level)
            elif attr in (
                "env",
                "maps",
                "plugins",
            ):
                lines += self.process_key_dict(attr, value, level, quote_key=False)

            elif attr == "projection":
                projection_comments = self.process_attribute_comment(comments, attr)
                lines += self.process_projection(
                    attr, value, level, projection_comments
                )
            elif attr in REPEATED_KEYS:
                lines += self.process_repeated_list(
                    attr, value, level, aligned_max_indent
                )
            elif attr == "points":
                lines += self.format_repeated_pair_list(attr, value, level)
            elif attr == "config":
                lines += self.process_config_dict(value, level)
            elif self.is_composite(value):
                lines += self._format(
                    value, level + 1
                )  # recursively add the child class
            else:
                # standard key value pair
                if not type_:
                    raise UnboundLocalError(
                        "The Mapfile object is missing a __type__ attribute"
                    )
                line = self.process_attribute(
                    type_, attr, value, level, aligned_max_indent
                )
                line += self.process_attribute_comment(comments, attr)
                lines.append(line)

        if not is_hidden:
            # close the container block with an END
            lines.append(self.add_end_line(level, 0, type_))

        return lines



# =================================================================
#
# Authors: Seth Girvin
#
# Copyright (c) 2023 Seth Girvin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from __future__ import annotations
from typing import Any
from itertools import zip_longest


def find(lst: list[dict], key: str, value: Any) -> dict | None:
    """
    Find an item in a list of dicts using a key and a value

    Parameters
    ----------

    list: list
        A list of composite dictionaries e.g. ``layers``, ``classes``
    key: string
        The key name to search each dictionary in the list
    key: value
        The value to search for

    Returns
    -------

    dict
        The first composite dictionary object with a key that matches the value

    Example
    -------

    To find the ``LAYER`` in a list of layers with ``NAME`` set to ``Layer2``::

        s = '''
        MAP
            LAYER
                NAME "Layer1"
                TYPE POLYGON
            END
            LAYER
                NAME "Layer2"
                TYPE POLYGON
                CLASS
                    NAME "Class1"
                    COLOR 0 0 -8
                END
            END
        END
        '''

        d = mappyfile.loads(s)
        cmp = mappyfile.find(d["layers"], "name", "Layer2")
        assert cmp["name"] == "Layer2"
    """
    return next((item for item in lst if item[key.lower()] == value), None)


def findall(lst: list[dict], key: str, value: Any) -> list[dict]:
    """
    Find all items in lst where key matches value.
    For example find all ``LAYER`` s in a ``MAP`` where ``GROUP`` equals ``VALUE``

    Parameters
    ----------

    list: list
        A list of composite dictionaries e.g. ``layers``, ``classes``
    key: string
        The key name to search each dictionary in the list
    key: value
        The value to search for

    Returns
    -------

    list
        A Python list containing the matching composite dictionaries

    Example
    -------

    To find all ``LAYER`` s with ``GROUP`` set to ``test``::

        s = '''
        MAP
            LAYER
                NAME "Layer1"
                TYPE POLYGON
                GROUP "test"
            END
            LAYER
                NAME "Layer2"
                TYPE POLYGON
                GROUP "test1"
            END
            LAYER
                NAME "Layer3"
                TYPE POLYGON
                GROUP "test2"
            END
            LAYER
                NAME "Layer4"
                TYPE POLYGON
                GROUP "test"
            END
        END
        '''

        d = mappyfile.loads(s)
        layers = mappyfile.findall(d["layers"], "group", "test")
        assert len(layers) == 2
    """
    return [item for item in lst if item[key.lower()] and item[key.lower()] in value]


def findunique(lst, key):
    """
    Find all unique key values for items in lst. If no
    items with the key are found an empty list is returned.

    Parameters
    ----------

    lst: list
         A list of composite dictionaries e.g. ``layers``, ``classes``
    key: string
        The key name to search each dictionary in the list

    Returns
    -------

    list
        A sorted Python list of unique keys in the list

    Example
    -------

    To find all ``GROUP`` values for ``CLASS`` in a ``LAYER``::

        s = '''
        LAYER
            CLASS
                GROUP "group1"
                NAME "Class1"
                COLOR 0 0 0
            END
            CLASS
                GROUP "group2"
                NAME "Class2"
                COLOR 0 0 0
            END
            CLASS
                GROUP "group1"
                NAME "Class3"
                COLOR 0 0 0
            END
        END
        '''

        d = mappyfile.loads(s)
        groups = mappyfile.findunique(d["classes"], "group")
        assert groups == ["group1", "group2"]
    """
    return sorted(
        set((item.get(key.lower(), None) for item in lst))
        - {
            None,
        }
    )


def findkey(d: dict, *keys: Any) -> dict:
    """
    Get a value from a dictionary based on a list of keys and/or list indexes.

    Parameters
    ----------

    d: dict
        A Python dictionary
    keys: list
        A list of key names, or list indexes

    Returns
    -------

    dict
        The composite dictionary object at the path specified by the keys

    Example
    -------

    To return the value of the first class of the first layer in a Mapfile::

        s = '''
        MAP
            LAYER
                NAME "Layer1"
                TYPE POLYGON
                CLASS
                    NAME "Class1"
                    COLOR 0 0 255
                END
            END
        END
        '''

        d = mappyfile.loads(s)

        pth = ["layers", 0, "classes", 0]
        cls1 = mappyfile.findkey(d, *pth)
        assert cls1["name"] == "Class1"
    """
    if keys:
        keys_list = list(keys)
        search_key = keys_list.pop(0)
        return findkey(d[search_key], *keys_list)

    return d


def update(d1: dict, d2: dict, overwrite: bool = True) -> dict:
    """
    Update dict d1 with properties from d2

    Note
    ----

    Allows deletion of objects with a special ``__delete__`` key
    For any list of dicts new items can be added when updating

    Parameters
    ----------

    d1: dict
        A Python dictionary
    d2: dict
        A Python dictionary that will be used to update any keys with the same name in d1
    overwrite: boolean
        If a key already exists in the dictionary should its value be overwritten

    Returns
    -------

    dict
        The updated dictionary

    """
    none_type = type(None)

    if d2.get("__delete__", False):
        return {}

    for k, v in d2.items():
        if isinstance(v, dict):
            if v.get("__delete__", False):
                # allow a __delete__ property to be set to delete objects
                del d1[k]
            else:
                d1[k] = update(d1.get(k, {}), v, overwrite)
        elif isinstance(v, (tuple, list)) and all(
            isinstance(li, (none_type, dict)) for li in v  # type: ignore
        ):
            # a list of dicts and/or none_type
            orig_list = d1.get(k, [])
            new_list = []
            pairs = list(zip_longest(orig_list, v, fillvalue=None))
            for orig_item, new_item in pairs:
                if orig_item is None:
                    orig_item = (
                        {}
                    )  # can't use {} for fillvalue as only one dict created/modified!
                if new_item is None:
                    new_item = {}

                if new_item.get("__delete__", False):
                    d = None  # orig_list.remove(orig_item) # remove the item to delete
                else:
                    d = update(orig_item, new_item, overwrite)

                if d is not None:
                    new_list.append(d)
            d1[k] = new_list
        else:
            if k in d1 and v == "__delete__":
                del d1[k]
            else:
                if overwrite is True or k not in d1:
                    d1[k] = v
    return d1


def dict_move_to_end(ordered_dict, key):
    ordered_dict.move_to_end(key)



# =================================================================
#
# Authors: Seth Girvin
#
# Copyright (c) 2020 Seth Girvin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from __future__ import annotations
import json
import os
from pathlib import Path
import platform
from collections import OrderedDict
import logging
from functools import lru_cache
import jsonschema
from mappyfile import dictutils
from referencing import Registry, Resource
from referencing.jsonschema import DRAFT202012
from typing import Any


log = logging.getLogger("mappyfile")


class Validator:
    def __init__(self):
        self.schemas = {}
        self.expanded_schemas = {}

    def get_schema_path(self, schemas_folder: str) -> str:
        """
        Return a file protocol URI e.g. file:///D:/mappyfile/mappyfile/schemas/ on Windows
        and file:////home/user/mappyfile/mappyfile/schemas/ on Linux
        """

        # replace any Windows path back slashes with forward slashes
        schemas_folder = schemas_folder.replace("\\", "/")

        # HACK Python 2.7 on Linux seems to remove the root slash
        # so add this back in
        if schemas_folder.startswith("/"):
            schemas_folder = "/" + schemas_folder

        host = ""
        root_schema_path = f"file://{host}/{schemas_folder}/"

        return root_schema_path

    @lru_cache(maxsize=None)
    def get_schemas_folder(self) -> str:
        base = Path(__file__).parent
        if platform.system() == "Windows":
            # keep drive letters (and avoid UNC paths)
            return str((base / "schemas").absolute())
        else:
            # resolve symlinks on Unix-like systems
            return str((base / "schemas").resolve())

    @lru_cache(maxsize=None)
    def get_schema_file(self, schema_name: str) -> str:
        file_extension = os.path.splitext(schema_name)[1]

        # when loading from the schema files the ref already includes .json
        if file_extension != ".json":
            schema_name += ".json"

        schemas_folder = self.get_schemas_folder()
        schema_file = os.path.join(schemas_folder, schema_name)

        if not os.path.isfile(schema_file):
            raise IOError(f"The file {schema_file} does not exist")

        return schema_file

    def get_json_from_file(self, schema_name: str):
        if schema_name not in self.schemas:
            schema_file = self.get_schema_file(schema_name)

            with open(schema_file, encoding="utf-8") as f:
                try:
                    jsn_schema = json.load(f)
                except ValueError as ex:
                    log.error("Could not load %s", schema_file)
                    raise ex

            # cache the schema contents to avoid re-reading the file a second time
            self.schemas[schema_name] = jsn_schema
        else:
            jsn_schema = self.schemas[schema_name]

        return jsn_schema

    def retrieve_from_filesystem(self, schema_name: str):
        jsn_schema = self.get_json_from_file(schema_name)
        return Resource.from_contents(jsn_schema, default_specification=DRAFT202012)

    def is_valid_for_version(self, d: dict, version: float) -> bool:
        """
        Check if the dict object has a metadata dict
        and if it does then check if the supplied
        version is valid against and minVersion or maxVersion
        values
        """
        if "metadata" in d:
            md = d["metadata"]
            min_version = md.get("minVersion", 0.0)
            max_version = md.get("maxVersion", 1000.0)
            if version < min_version or version > max_version:
                return False

        return True

    def get_versioned_schema(
        self, version: float | None = None, schema_name: str = "map"
    ) -> dict:
        """
        Get a fully expanded JSON schema for a specific MapServer
        version. Optionally provide a schema_name to return an expanded
        subsection of the full schema.
        """
        jsn_schema = self.get_expanded_schema(schema_name, version)

        if version:
            # remove any properties based on minVersion and maxVersion
            properties = jsn_schema["properties"]
            jsn_schema["properties"] = self.get_versioned_properties(
                properties, version
            )

        return jsn_schema

    def get_versioned_properties(self, properties: dict, version: float) -> dict:
        """
        For a dict object recursively check each child object
        to see if it is valid for the supplied version
        """
        keys_copy = list(properties.keys())

        for key in keys_copy:
            v = properties[key]
            if isinstance(v, dict):
                if self.is_valid_for_version(v, version) is False:
                    del properties[key]
                self.get_versioned_properties(v, version)
            elif isinstance(v, list):
                valid_list = []
                for props in v:
                    if isinstance(props, dict):
                        if self.is_valid_for_version(props, version) is True:
                            valid_list.append(props)
                    else:
                        valid_list.append(props)
                properties[key] = valid_list

        return properties

    def get_schema_validator(self, schema_name: str) -> jsonschema.Draft202012Validator:
        """
        Create a schema validator
        """

        jsn_schema = self.get_json_from_file(schema_name)

        registry: Registry = Registry(retrieve=self.retrieve_from_filesystem)  # type: ignore

        validator = jsonschema.Draft202012Validator(schema=jsn_schema, registry=registry)  # type: ignore
        # validator.check_schema(jsn_schema) # check schema is valid - commented out as for testing only

        return validator

    def convert_lowercase(self, x: Any) -> Any:
        if isinstance(x, list):
            return [self.convert_lowercase(v) for v in x]

        if isinstance(x, dict):
            return OrderedDict(
                (k.lower(), self.convert_lowercase(v)) for k, v in x.items()
            )

        if isinstance(x, (str, bytes)):
            x = x.lower()

        return x

    def create_message(
        self, rootdict: dict, path: list[Any], error, add_comments: bool
    ) -> dict:
        """
        Add a validation comment to the dictionary
        path is the path to the error object, it can be empty if the error is in the root object
        http://python-jsonschema.readthedocs.io/en/latest/errors/#jsonschema.exceptions.ValidationError.absolute_path
        It can also reference an object in a list e.g. [u'layers', 0]

        Unfortunately it is not currently possible to get the name of the failing property from the
        JSONSchema error object, even though it is in the error message.
        See https://github.com/Julian/jsonschema/issues/119
        """

        if not path:
            # error applies to the root type
            d = rootdict
            key = d["__type__"]
        elif isinstance(path[-1], int):
            # the error is on an object in a list
            d = dictutils.findkey(rootdict, *path)
            key = d["__type__"]
        else:
            key = path[-1]
            d = dictutils.findkey(rootdict, *path[:-1])

        error_message = f"ERROR: Invalid value in {key.upper()}"

        # add a comment to the dict structure

        if add_comments:
            if "__comments__" not in d:
                d["__comments__"] = OrderedDict()

            d["__comments__"][key] = f"# {error_message}"

        error_dict = {"error": error.message, "message": error_message}

        # add in details of the error line, when Mapfile was parsed to
        # include position details

        if "__position__" in d:
            if not path or key not in d["__position__"]:
                # position for the root object is stored in the root of the dict
                pd = d["__position__"]
            else:
                pd = d["__position__"][key]

            error_dict["line"] = pd.get("line")
            error_dict["column"] = pd.get("column")

        return error_dict

    def get_error_messages(
        self, d: dict, errors: list[Any], add_comments: bool
    ) -> list[dict]:
        error_messages = []

        for error in errors:
            pth = error.absolute_path
            pth = list(pth)  # convert deque to list
            em = self.create_message(d, pth, error, add_comments)
            error_messages.append(em)

        return error_messages

    def _get_errors(
        self, d: dict, validator: jsonschema.Draft202012Validator, add_comments: bool
    ) -> list[dict]:
        lowercase_dict = self.convert_lowercase(d)
        jsn = json.loads(json.dumps(lowercase_dict), object_pairs_hook=OrderedDict)

        errors = list(validator.iter_errors(jsn))
        error_messages = self.get_error_messages(d, errors, add_comments)

        return error_messages

    def validate(
        self,
        value: Any,
        add_comments: bool = False,
        schema_name: str = "map",
        version: float | None = None,
    ):
        """
        verbose - also return the jsonschema error details
        """
        if version:
            jsn_schema = self.get_versioned_schema(version, schema_name)
            validator = jsonschema.Draft202012Validator(schema=jsn_schema)
        else:
            validator = self.get_schema_validator(schema_name)

        error_messages = []

        if isinstance(value, list):
            for d in value:
                error_messages += self._get_errors(d, validator, add_comments)
        else:
            error_messages = self._get_errors(value, validator, add_comments)

        return error_messages

    def expand_refs(self, schema: str):
        if isinstance(schema, dict):
            if "$ref" in schema:
                schema_name = schema["$ref"]
                ref_schema = self.get_json_from_file(schema_name)
                return self.expand_refs(ref_schema)

            return {k: self.expand_refs(v) for k, v in schema.items()}

        elif isinstance(schema, list):
            return [self.expand_refs(item) for item in schema]

        return schema

    def get_expanded_schema(
        self, schema_name: str, version: float | None = None
    ) -> dict:
        """
        Return a schema file with all $ref properties expanded
        """

        # ensure we cache per version to handle validator reuse
        if version is not None:
            cache_schema_name = schema_name + str(version)
        else:
            cache_schema_name = schema_name

        if cache_schema_name not in self.expanded_schemas:
            root_schema = self.get_json_from_file(schema_name)
            jsn_schema = self.expand_refs(root_schema)

            # cache the schema for future use
            self.expanded_schemas[cache_schema_name] = jsn_schema
        else:
            jsn_schema = self.expanded_schemas[cache_schema_name]

        return jsn_schema



# =================================================================
#
# Authors: Seth Girvin, Erez Shinan
#
# Copyright (c) 2020 Seth Girvin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from __future__ import annotations
import os
import logging
from io import open
from lark import Lark, ParseError, Tree, UnexpectedInput
from typing import Any, IO


log = logging.getLogger("mappyfile")


def string_to_boolean(string_value: str):
    return string_value.lower() not in ("false", "no", "0", "off")


use_cython = string_to_boolean(os.environ.get("MAPPYFILE_USE_CYTHON", "True"))
lark_cython = None

if use_cython:
    try:
        import lark_cython  # type: ignore
    except ImportError:
        pass


SYMBOL_ATTRIBUTES = {
    "ANCHORPOINT",
    "ANTIALIAS",
    "FILLED",
    "FONT",
    "IMAGE",
    "NAME",
    "COLOR",
    "TYPE",
    "CHARACTER",
    "POINTS",
    "TRANSPARENT",
}


class Parser:
    def __init__(
        self, expand_includes: bool = True, include_comments: bool = False, **kwargs
    ):
        self.expand_includes = expand_includes
        self.include_comments = include_comments
        self._comments: list[Any] = []
        self.lalr = self._create_lalr_parser()
        self.kwargs = kwargs

    def _create_lalr_parser(self) -> Any:
        extra_args = {}

        if lark_cython:
            extra_args["_plugins"] = lark_cython.plugins

        if self.include_comments:
            callbacks = {
                "COMMENT": self._comments.append,
                "CCOMMENT": self._comments.append,
            }
            extra_args.update(
                {"propagate_positions": True, "lexer_callbacks": callbacks}
            )

        return Lark.open("mapfile.lark", rel_to=__file__, parser="lalr", **extra_args)

    def _get_include_filename(self, line: str) -> str:
        if "#" in line:
            # remove any comments on the same line
            line = line.split("#")[0]

        include_pairs = line.split()
        if len(include_pairs) > 2:
            log.warning(
                "Multiple include files have been found on the same line. "
                "Only the first will be used. "
            )
        inc_file_path = include_pairs[1]

        return inc_file_path.strip("'").strip('"')

    def load_includes(
        self, text: str, fn: str | None = None, _nested_includes: int = 0
    ) -> str:
        # Per default use working directory of the process
        if fn is None:
            fn = os.getcwd() + os.sep

        lines = text.split("\n")
        includes = {}
        for idx, l in enumerate(lines):
            if l.strip().lower().startswith("include"):
                if _nested_includes == 5:
                    raise ValueError("Maximum nested include exceeded! (MaxNested=5)")

                inc_file_path = self._get_include_filename(l)

                if not os.path.isabs(inc_file_path):
                    inc_file_path = os.path.abspath(
                        os.path.join(os.path.dirname(fn), inc_file_path)
                    )
                try:
                    include_text = self.open_file(inc_file_path)
                except IOError as ex:
                    log.warning(
                        "Include file '%s' not found in '%s'", inc_file_path, fn
                    )
                    raise ex
                # recursively load any further includes
                includes[idx] = self.load_includes(
                    include_text, fn=fn, _nested_includes=_nested_includes + 1
                )

        for idx, txt in includes.items():
            lines.pop(idx)  # remove the original include
            lines.insert(idx, txt)
        return "\n".join(lines)

    def _assign_comments(self, _tree: Any) -> None:
        for node in _tree.children:
            if not isinstance(node, Tree):
                continue

            if not hasattr(node.meta, "line"):
                continue

            line = node.meta.line

            # when we encounter a new type that can have associated comments
            # assign all comments up to that point in the Mapfile to the node
            # for metadata we want to assign comments to the string_pair
            if node.data in (
                "composite",
                "attr",
                "projection",
                "string_pair",
                "config_attr",
            ):
                # for projection blocks capture any comments within the block

                if node.data in ("projection"):
                    line = node.meta.end_line
                line_numbers = list(sorted(self.comments_dict.keys()))
                comments = []

                for line_number in line_numbers:
                    if line_number <= line:
                        comments.append(self.comments_dict.pop(line_number))

                if comments:
                    node.meta.comments = comments  # type: ignore

            if isinstance(node, Tree):
                self._assign_comments(node)

    def load(self, fp: IO[str]) -> Any:
        text = fp.read()
        if hasattr(fp, "name"):
            fn = (
                fp.name
            )  # name is a read-only attribute and may not be present on all file-like objects.
        else:
            fn = None
        return self.parse(text, fn)

    def open_file(self, fn: str):
        try:
            with open(fn, "r", encoding="utf-8") as f:
                return f.read()
        except UnicodeDecodeError as ex:
            log.debug(ex)
            log.error(
                "Please check the encoding for %s. All Mapfiles should be in utf-8 format.",
                fn,
            )
            raise

    def parse_file(self, fn: str) -> Any:
        text = self.open_file(fn)
        return self.parse(text, fn=fn)

    def parse(self, text: str, fn: str | None = None) -> Any:
        """
        Parse the Mapfile
        """

        if text.strip() == "":
            raise ValueError(f"The file {fn} is empty")

        if self.expand_includes:
            text = self.load_includes(text, fn=fn)

        try:
            self._comments[:] = []  # clear any comments from a previous parse
            ip = self.lalr.parse_interactive(text)
            for t in ip.iter_parse():
                if t.type == "UNQUOTED_STRING":
                    # Unquoted strings after SYMBOL can only be values, not attributes
                    if (
                        ip.parser_state.value_stack[-1] == "SYMBOL"
                        and t.value.upper() not in SYMBOL_ATTRIBUTES
                    ):
                        t.type = "UNQUOTED_STRING_VALUE"
                elif t.type == "GRID":
                    # Unquoted 'GRID' coming after NAME is always a value, not a composite type
                    if ip.parser_state.value_stack[-1] == "NAME":
                        t.type = "UNQUOTED_STRING_VALUE"

            tree = ip.resume_parse()
            if self.include_comments:
                self.comments_dict = {}
                # create a dictionary using line numbers as keys, and comments as values
                for c in self._comments:
                    self.comments_dict[c.line] = c.value.strip()
                self._assign_comments(tree)

            return tree
        except (ParseError, UnexpectedInput) as ex:
            if fn:
                log.error("Parsing of %s unsuccessful", fn)
            else:
                log.error("Parsing of Mapfile unsuccessful")
            log.info(ex)
            raise
